@@ -26,6 +26,7 @@
 #include <gnuradio/io_signature.h>
 #include "source_impl.h"
 #include <SoapySDR/Formats.h>
+#include <cmath>
 
 const pmt::pmt_t CMD_CHAN_KEY = pmt::mp("chan");
 const pmt::pmt_t CMD_FREQ_KEY = pmt::mp("freq");
@@ -37,37 +38,48 @@ const pmt::pmt_t CMD_BW_KEY = pmt::mp("bw");
 namespace gr {
 namespace soapy {
 source::sptr
-source::make(size_t nchan, const std::string device,
-             const std::string args, double sampling_rate,
-             const std::string type)
+source::make(size_t nchan, const std::string &device,
+             const std::string &dev_args,
+             const std::string &stream_args,
+             const std::vector<std::string> &tune_args,
+             const std::vector<std::string> &other_settings,
+             double sampling_rate,
+             const std::string &type)
 {
   return gnuradio::get_initial_sptr(
-           new source_impl(nchan, device, args, sampling_rate, type));
+           new source_impl(nchan, device, dev_args, stream_args,
+                           tune_args, other_settings, sampling_rate, type));
 }
 
 /*
  * The private constructor
  */
-source_impl::source_impl(size_t nchan, const std::string device,
-                         const std::string args, double sampling_rate,
-                         const std::string type) :
+source_impl::source_impl(size_t nchan, const std::string &device,
+                         const std::string &dev_args,
+                         const std::string &stream_args,
+                         const std::vector<std::string> &tune_args,
+                         const std::vector<std::string> &other_settings,
+                         double sampling_rate,
+                         const std::string &type) :
   gr::sync_block("soapy::source", gr::io_signature::make(0, 0, 0),
                  args_to_io_sig(type, nchan)),
-  d_dev_str(device),
   /*
    * Swig does not guarantee C++ destructors are called from python,
    *  so recommends against relying on them for cleanup.
    */
   d_stopped(true),
   d_mtu(0),
-  d_sampling_rate(sampling_rate),
   d_nchan(nchan)
 {
-  /*
-   * Get the device in a key, value way. The device string should contain the
-   * driver=xxxx and therefore we can identify the device type
-   */
-  SoapySDR::Kwargs kwargs = SoapySDR::KwargsFromString(device);
+  if (nchan != tune_args.size()) {
+    std::string msg = name() +  ": Wrong number of channels and tune arguments";
+    throw std::invalid_argument(msg);
+  }
+
+  if (nchan != tune_args.size()) {
+    std::string msg = name() +  ": Wrong number of channels and settings";
+    throw std::invalid_argument(msg);
+  }
 
   std::string stype;
   if (type == "fc32") {
@@ -84,7 +96,20 @@ source_impl::source_impl(size_t nchan, const std::string device,
     throw std::invalid_argument(msg);
   }
 
-  makeDevice(device);
+  SoapySDR::Kwargs dev_kwargs =
+    SoapySDR::KwargsFromString(device +  ", " + dev_args);
+  d_device = SoapySDR::Device::make(dev_kwargs);
+  d_stopped = false;
+
+  /* Some of the arguments maybe device settings, so we need to re-apply them*/
+  SoapySDR::ArgInfoList supported_settings = d_device->getSettingInfo();
+  for (const std::pair<std::string, std::string> &i : dev_kwargs) {
+    for (SoapySDR::ArgInfo &arg : supported_settings) {
+      if (arg.key == i.first) {
+        d_device->writeSetting(i.first, i.second);
+      }
+    }
+  }
 
   if (d_nchan > d_device->getNumChannels(SOAPY_SDR_RX)) {
     std::string msg = name() +  ": Unsupported number of channels. Only  "
@@ -95,65 +120,129 @@ source_impl::source_impl(size_t nchan, const std::string device,
 
   std::vector<size_t> channs;
   channs.resize(d_nchan);
+  bool accept_samp_rate = false;
   for (size_t i = 0; i < d_nchan; i++) {
     channs[i] = i;
 
-    SoapySDR::RangeList sampRange = d_device->getSampleRateRange(SOAPY_SDR_RX, i);
+    SoapySDR::RangeList sps_range = d_device->getSampleRateRange(SOAPY_SDR_RX, i);
 
-    double minRate = sampRange.front().minimum();
-    double maxRate = sampRange.back().maximum();
+    for (SoapySDR::Range &range : sps_range) {
+      double min_sps = range.minimum();
+      double max_sps = range.maximum();
 
-    // for some reason the airspy provides them backwards
-    if (minRate > maxRate) {
-      std::swap(minRate, maxRate);
+      if ((sampling_rate < min_sps) || (sampling_rate > max_sps)) {
+        continue;
+      }
+      else {
+        accept_samp_rate = true;
+        break;
+      }
     }
+    if (!accept_samp_rate) {
+      std::string msg = name() +
+                        ": Unsupported sample rate.  Rate must be in the range";
+      for (SoapySDR::Range &range : sps_range) {
+        double min_sps = range.minimum();
+        double max_sps = range.maximum();
+        msg += " [" + std::to_string(min_sps) +
+               ", " + std::to_string(max_sps) + "]";
+      }
+      throw std::invalid_argument(msg);
+    }
+    set_sample_rate(i, sampling_rate);
+  }
 
-    if (kwargs["driver"] != "airspy") {
-      if ((d_sampling_rate < minRate) || (d_sampling_rate > maxRate)) {
-        std::string msg = name() + ": Unsupported sample rate.  Rate must be between "
-                          + std::to_string(minRate)
-                          + " and " + std::to_string(maxRate);
+  /*
+   * Check for the validity of the specified stream arguments on all enabled
+   * channels
+   */
+  for (size_t chan : channs) {
+    SoapySDR::ArgInfoList supported_args =
+      d_device->getStreamArgsInfo(SOAPY_SDR_RX, chan);
+    SoapySDR::Kwargs stream_kwargs = SoapySDR::KwargsFromString(stream_args);
+
+    for (const std::pair<std::string, std::string> &i : stream_kwargs) {
+      bool found = false;
+      for (SoapySDR::ArgInfo &arg : supported_args) {
+        if (arg.key == i.first) {
+          found = true;
+          break;
+        }
+      }
+      if (!found) {
+        std::string msg = name() + ": Unsupported stream argument " + i.first
+                          + " for channel " + std::to_string(chan);
         throw std::invalid_argument(msg);
       }
     }
-    else {
-      if (((long)d_sampling_rate != 2500000) && ((long)d_sampling_rate != 10000000)) {
-        throw std::invalid_argument(name() + ": Airspy only supports 2.5 MSPS "
-                                    "and 10 MSPS rates.  Requested "
-                                    + std::to_string((long)d_sampling_rate));
-
-      }
-    }
-    set_sample_rate(i, d_sampling_rate);
   }
-
-  /* Apply to all streams the supported args */
-  SoapySDR::ArgInfoList supported_args = d_device->getStreamArgsInfo(SOAPY_SDR_RX,
-                                         0);
-  SoapySDR::Kwargs dev_args = SoapySDR::KwargsFromString(args);
-  SoapySDR::Kwargs stream_args;
-  for (const SoapySDR::ArgInfo &i : supported_args) {
-    const SoapySDR::Kwargs::const_iterator iter = dev_args.find(i.key);
-    if (iter != dev_args.end()) {
-      stream_args[iter->first] = iter->second;
-      /*
-       * This was a stream argument. Erase it so it cannot be applied later
-       * as device argument
-       */
-      dev_args.erase(i.key);
-    }
-  }
-
-  d_stream = d_device->setupStream(SOAPY_SDR_RX, stype, channs, stream_args);
+  d_stream = d_device->setupStream(SOAPY_SDR_RX, stype, channs,
+                                   SoapySDR::KwargsFromString(stream_args));
   d_mtu = d_device->getStreamMTU(d_stream);
 
-  /*
-   * Apply device settings to all enabled channels.
-   * These should be any settings that were not a stream argument
-   */
+  /* Apply tuning arguments on the enabled channels */
   for (size_t chan : channs) {
-    for (const std::pair<std::string, std::string> &iter : dev_args) {
-      d_device->writeSetting(SOAPY_SDR_RX, chan, iter.first, iter.second);
+    SoapySDR::ArgInfoList supported_args =
+      d_device->getFrequencyArgsInfo(SOAPY_SDR_RX, chan);
+    SoapySDR::Kwargs tune_kwargs = SoapySDR::KwargsFromString(tune_args[chan]);
+
+    for (const std::pair<std::string, std::string> &i : tune_kwargs) {
+      bool found = false;
+      for (SoapySDR::ArgInfo &arg : supported_args) {
+        if (arg.key == i.first) {
+          found = true;
+          break;
+        }
+      }
+      if (!found) {
+        std::string msg = name() + ": Unsupported tune argument " + i.first
+                          + " for channel " + std::to_string(chan);
+        throw std::invalid_argument(msg);
+      }
+    }
+    d_tune_args.push_back(tune_kwargs);
+  }
+
+  /* Apply any other channel specific setting */
+  for (size_t chan : channs) {
+    SoapySDR::ArgInfoList supported_settings =
+      d_device->getSettingInfo(SOAPY_SDR_RX, chan);
+    SoapySDR::Kwargs settings_kwargs =
+      SoapySDR::KwargsFromString(other_settings[chan]);
+
+    /*
+     * Before applying the setting, make a last check if the setting
+     * is a valid gain
+     */
+    std::vector<std::string> gains = d_device->listGains(SOAPY_SDR_RX, chan);
+    for (const std::string &i : gains) {
+      SoapySDR::Kwargs::iterator iter = settings_kwargs.find(i);
+      if (iter != settings_kwargs.end()) {
+        d_device->setGain(SOAPY_SDR_RX, chan, i, std::stod(iter->second));
+        /*
+         * As only valid settings are applied, we remove it so it does not
+         * recognized as an invalid setting
+         */
+        settings_kwargs.erase(iter);
+      }
+    }
+
+    for (const std::pair<std::string, std::string> &i : settings_kwargs) {
+      bool found = false;
+      for (SoapySDR::ArgInfo &arg : supported_settings) {
+        if (arg.key == i.first) {
+          found = true;
+          break;
+        }
+      }
+      if (found) {
+        d_device->writeSetting(SOAPY_SDR_RX, chan, i.first, i.second);
+      }
+      else {
+        std::string msg = name() + ": Unsupported setting " + i.first
+                          + " for channel " + std::to_string(chan);
+        throw std::invalid_argument(msg);
+      }
     }
   }
 
@@ -177,11 +266,13 @@ source_impl::source_impl(size_t nchan, const std::string device,
     CMD_ANTENNA_KEY,
     boost::bind(&source_impl::cmd_handler_antenna, this, _1, _2));
 
-  /* GNU Radio stream engine is very efficient when the buffers are a power of 2*/
-  set_output_multiple(1024);
-
   /* This limits each work invocation to MTU transfers */
-  set_max_noutput_items(d_mtu);
+  if (d_mtu > 0) {
+    set_max_noutput_items(d_mtu);
+  }
+  else {
+    set_max_noutput_items(1024);
+  }
 }
 
 bool
@@ -196,7 +287,7 @@ source_impl::stop()
 {
   if (!d_stopped) {
     d_device->closeStream(d_stream);
-    unmakeDevice(d_device);
+    SoapySDR::Device::unmake(d_device);
     d_stopped = true;
   }
   return true;
@@ -222,43 +313,16 @@ source_impl::register_msg_cmd_handler(const pmt::pmt_t &cmd,
   d_cmd_handlers[cmd] = handler;
 }
 
-void
-source_impl::makeDevice(const std::string &argStr)
-{
-  d_device = SoapySDR::Device::make(argStr);
-  d_stopped = false;
-}
-
-void
-source_impl::unmakeDevice(SoapySDR::Device *dev)
-{
-  SoapySDR::Device::unmake(dev);
-}
-
-bool source_impl::hasDCOffset(int channel)
-{
-  return d_device->hasDCOffset(SOAPY_SDR_RX, channel);
-}
-
-bool source_impl::hasIQBalance(int channel)
-{
-  return d_device->hasIQBalance(SOAPY_SDR_RX, channel);
-}
-
-bool source_impl::hasFrequencyCorrection(int channel)
-{
-  return d_device->hasFrequencyCorrection(SOAPY_SDR_RX, channel);
-}
-
 void
 source_impl::set_frequency(size_t channel, double frequency)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
   }
 
-  d_device->setFrequency(SOAPY_SDR_RX, channel, frequency);
-  d_frequency = d_device->getFrequency(SOAPY_SDR_RX, channel);
+  d_device->setFrequency(SOAPY_SDR_RX, channel, frequency, d_tune_args[channel]);
 }
 
 void
@@ -266,7 +330,22 @@ source_impl::set_frequency(size_t channel, const std::string &name,
                            double frequency)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
+  }
+
+  std::vector<std::string> freqs = d_device->listFrequencies(SOAPY_SDR_RX,
+                                   channel);
+
+  if (std::find(freqs.begin(), freqs.end(), name) == freqs.end()) {
+    if (std::fpclassify(std::abs(frequency)) == FP_ZERO) {
+      return;
+    }
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " does not support frequency " + name
+                      + ". Consider setting this frequency to 0 to bypass this error";
+    throw std::invalid_argument(msg);
   }
 
   d_device->setFrequency(SOAPY_SDR_RX, channel, name, frequency);
@@ -287,8 +366,11 @@ void
 source_impl::set_gain(size_t channel, float gain)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
   }
+
   SoapySDR::Range rGain = d_device->getGainRange(SOAPY_SDR_RX, channel);
 
   if (gain < rGain.minimum() || gain > rGain.maximum()) {
@@ -299,36 +381,19 @@ source_impl::set_gain(size_t channel, float gain)
   }
 
   d_device->setGain(SOAPY_SDR_RX, channel, gain);
-  d_gain = d_device->getGain(SOAPY_SDR_RX, channel);
 }
 
 void
-source_impl::set_overall_gain(size_t channel, float gain, bool manual_mode)
+source_impl::set_gain(size_t channel, const std::string name, float gain)
 {
-  if (channel >= d_nchan || manual_mode) {
-    return;
-  }
-  set_gain(channel, gain);
-}
-
-void
-source_impl::set_gain(size_t channel, const std::string name, float gain,
-                      bool manual_mode)
-{
-  /*
-   * This setter is for manual mode gain. There is a known limitation of the
-   * GRC and the yaml runtime evaluation, so we need to skip this setting
-   * if the mode is auto gain
-   */
-  if (channel >= d_nchan || !manual_mode) {
+  if (channel >= d_nchan) {
     return;
   }
 
   if (!gain_available(channel, name)) {
-    GR_LOG_WARN(d_logger,
-                boost::format("Unknown %s gain setting "
-                              "for channel %zu") % name % channel);
-    return;
+    std::string msg = this->name() + ": Unknown gain " + name + " for channel "
+                      + std::to_string(channel);
+    throw std::invalid_argument(msg);
   }
 
   SoapySDR::Range rGain = d_device->getGainRange(SOAPY_SDR_RX, channel, name);
@@ -343,27 +408,36 @@ source_impl::set_gain(size_t channel, const std::string name, float gain,
 }
 
 void
-source_impl::set_gain_mode(size_t channel, bool gain_auto_mode)
+source_impl::set_agc(size_t channel, bool enable)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
   }
 
-  d_device->setGainMode(SOAPY_SDR_RX, channel, gain_auto_mode);
+  if (!d_device->hasGainMode(SOAPY_SDR_RX, channel) && enable) {
+    std::string msg = name() + ": Channel " + std::to_string(channel)
+                      + " does not support AGC";
+    throw std::invalid_argument(msg);
+  }
+  d_device->setGainMode(SOAPY_SDR_RX, channel, enable);
 }
 
 void
 source_impl::set_sample_rate(size_t channel, double sample_rate)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
   }
 
   d_device->setSampleRate(SOAPY_SDR_RX, channel, sample_rate);
-  d_sampling_rate = sample_rate;
 }
 
-std::vector<std::string> source_impl::listAntennas(int channel)
+std::vector<std::string>
+source_impl::get_antennas(int channel)
 {
   if ((size_t)channel >= d_nchan) {
     return std::vector<std::string>();
@@ -376,73 +450,89 @@ void
 source_impl::set_bandwidth(size_t channel, double bandwidth)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
   }
 
   d_device->setBandwidth(SOAPY_SDR_RX, channel, bandwidth);
-  d_bandwidth = bandwidth;
 }
 
 void
 source_impl::set_antenna(const size_t channel, const std::string &name)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
   }
 
-  std::vector<std::string> antennaList = d_device->listAntennas(SOAPY_SDR_RX,
-                                         channel);
+  std::vector<std::string> antennas = d_device->listAntennas(SOAPY_SDR_RX,
+                                      channel);
 
-  if (antennaList.size() > 0) {
-    if (std::find(antennaList.begin(), antennaList.end(),
-                  name) == antennaList.end()) {
-      GR_LOG_WARN(d_logger,
-                  boost::format("Antenna name %s not supported.") % name);
+  if (antennas.size() > 0) {
+    if (std::find(antennas.begin(), antennas.end(),
+                  name) == antennas.end()) {
+      std::string msg = this->name() + ": Antenna " + name + " at channel "
+                        + std::to_string(channel) + " is not supported. "
+                        + "Available antennas are: [";
+      for (std::string i : antennas) {
+        msg += i + ", ";
+      }
+      msg.erase(msg.length() - 2, 2);
+      msg += "]";
+      throw std::invalid_argument(msg);
       return;
     }
   }
 
   d_device->setAntenna(SOAPY_SDR_RX, channel, name);
-  d_antenna = name;
 }
 
 void
-source_impl::set_dc_offset(size_t channel, gr_complexd dc_offset,
-                           bool dc_offset_auto_mode)
+source_impl::set_dc_offset(size_t channel, gr_complexd dc_offset)
 {
   if (channel >= d_nchan) {
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
+  }
+
+  const bool supported = d_device->hasDCOffset(SOAPY_SDR_RX, channel);
+  /*
+   * In case of 0 DC offset, the method exits to avoid failing in case of the
+   * device does not supports DC offset
+   */
+  if (std::fpclassify(std::norm(dc_offset)) == FP_ZERO) {
+    if (supported && !d_device->getDCOffsetMode(SOAPY_SDR_RX, channel)) {
+      d_device->setDCOffset(SOAPY_SDR_RX, channel, dc_offset);
+    }
     return;
   }
 
-  if (!hasDCOffset(channel)) {
-    return;
-  }
-
-  /* If DC Correction is supported but automatic mode is not set DC correction */
-  if (!dc_offset_auto_mode
-      && d_device->hasDCOffset(SOAPY_SDR_RX, channel)) {
-    d_device->setDCOffset(SOAPY_SDR_RX, channel, dc_offset);
-    d_dc_offset = dc_offset;
+  if (!supported) {
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " does not support DC offset setting";
+    throw std::invalid_argument(msg);
   }
+  d_device->setDCOffset(SOAPY_SDR_RX, channel, dc_offset);
 }
 
 void
-source_impl::set_dc_offset_mode(size_t channel, bool dc_offset_auto_mode)
+source_impl::set_dc_removal(size_t channel, bool automatic)
 {
   if (channel >= d_nchan) {
-    return;
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
   }
 
-  if (!hasDCOffset(channel)) {
-    return;
-  }
-
-  /* If user specifies automatic DC Correction and is supported activate it */
-  if (dc_offset_auto_mode
-      && d_device->hasDCOffsetMode(SOAPY_SDR_RX, channel)) {
-    d_device->setDCOffsetMode(SOAPY_SDR_RX, channel, dc_offset_auto_mode);
-    d_dc_offset_auto_mode = dc_offset_auto_mode;
+  if (!d_device->hasDCOffsetMode(SOAPY_SDR_RX, channel) && automatic) {
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " does not support automatic DC removal setting";
+    throw std::invalid_argument(msg);
   }
+  d_device->setDCOffsetMode(SOAPY_SDR_RX, channel, automatic);
 }
 
 void
@@ -450,51 +540,70 @@ source_impl::set_frequency_correction(size_t channel,
                                       double freq_correction)
 {
   if (channel >= d_nchan) {
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
+  }
+
+  const bool supported = d_device->hasFrequencyCorrection(SOAPY_SDR_RX, channel);
+  /*
+   * In case of 0 frequency correction, the method exits to avoid failing
+   * in case of the device does not supports DC offset
+   */
+  if (std::fpclassify(std::abs(freq_correction)) == FP_ZERO) {
+    if (supported) {
+      d_device->setFrequencyCorrection(SOAPY_SDR_RX, channel, freq_correction);
+    }
     return;
   }
 
-  if (!hasFrequencyCorrection(channel)) {
-    return;
-  }
-
-  /* If the device supports Frequency correction set value */
-  if (d_device->hasFrequencyCorrection(SOAPY_SDR_RX, channel)) {
-    d_device->setFrequencyCorrection(SOAPY_SDR_RX, channel,
-                                     freq_correction);
-    d_frequency_correction = freq_correction;
+  if (!supported) {
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " does not support frequency correction setting";
+    throw std::invalid_argument(msg);
   }
+  d_device->setFrequencyCorrection(SOAPY_SDR_RX, channel, freq_correction);
 }
 
 void
 source_impl::set_iq_balance(size_t channel, gr_complexd iq_balance)
 {
   if (channel >= d_nchan) {
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " is not activated.";
+    throw std::invalid_argument(msg);
+  }
+
+  const bool supported = d_device->hasIQBalance(SOAPY_SDR_RX, channel);
+  /*
+   * In case of 0 frequency correction, the method exits to avoid failing
+   * in case of the device does not supports DC offset
+   */
+  if (std::fpclassify(std::norm(iq_balance)) == FP_ZERO) {
+    if (supported) {
+      d_device->setIQBalance(SOAPY_SDR_RX, channel, iq_balance);
+    }
     return;
   }
 
-  if (!hasIQBalance(channel)) {
-    return;
-  }
-
-  /* If the device supports IQ blance correction set value */
-  if (d_device->hasIQBalance(SOAPY_SDR_RX, channel)) {
-    d_device->setIQBalance(SOAPY_SDR_RX, channel, iq_balance);
-    d_iq_balance = iq_balance;
+  if (!supported) {
+    std::string msg = this->name() + ": Channel " + std::to_string(channel)
+                      + " does not support IQ imbalance correction setting";
+    throw std::invalid_argument(msg);
   }
+  d_device->setIQBalance(SOAPY_SDR_RX, channel, iq_balance);
 }
 
 void
 source_impl::set_master_clock_rate(double clock_rate)
 {
   d_device->setMasterClockRate(clock_rate);
-  d_clock_rate = clock_rate;
 }
 
 void
 source_impl::set_clock_source(const std::string &clock_source)
 {
   d_device->setClockSource(clock_source);
-  d_clock_source = clock_source;
 }
 
 void
@@ -617,7 +726,7 @@ source_impl::work(int noutput_items,
     int read = d_device->readStream(d_stream, output_items.data(), noutput_items,
                                     flags, time_ns);
     boost::this_thread::restore_interruption restore_interrupt(disable_interrupt);
-    if (read > 0) {
+    if (read >= 0) {
       return read;
     }
 
